;;;; bloom-filter.scm
;;;; Kon Lovett, Jun '06

;; Issues
;;
;; - Not optimal for 64-bit machines.
;;
;; - Note the use of "in-place" list mutators, '(*! ...). Be vewy careful.
;;
;; - Although K may be lower the actual number of hash functions
;; are generated anyway. It is up to the caller to optimize.
;;
;; - Should unavailable message-digests be an error, rather than a warning?
;;
;; - No ability to add message-digest-primitives at runtime.

(module bloom-filter

  (;export
    optimum-k
    optimum-m
    p-random-one-bit
    p-false-positive
    desired-m
    actual-k
    make-bloom-filter
    bloom-filter? check-bloom-filter error-bloom-filter
    bloom-filter-n
    bloom-filter-m
    bloom-filter-k
    bloom-filter-p-false-positive
    bloom-filter-set!
    bloom-filter-exists?)

  (import scheme chicken)

  (use
    srfi-1 srfi-13
    numeric-macros moremacros
    iset message-digest
    record-variants type-checks hash-utils)

;;;

(include "mathh-constants")

(define-constant unsigned-integer32-size 4)

;;; Record Type

(define-record-type-variant bloom-filter (unsafe unchecked inline)
	(%make-bloom-filter n m k bits hashes)
	%bloom-filter?
	(n %bloom-filter-n %bloom-filter-n-set!)
	(m %bloom-filter-m)
	(k %bloom-filter-k)
	(bits %bloom-filter-bits %bloom-filter-bits-set!)
	(hashes %bloom-filter-hashes) )

;;; Support

(define (string->substring-list/shared str chunk-size #!optional (start 0) (end (##sys#size str)))
  (let* ((rem (remainder (- end start) chunk-size))
         (len (- end rem))
         (sublst
          (let loop ((pos start) (lst '()))
            (if (>= pos len) (reverse! lst)
                (let ((npos (+ pos chunk-size)))
                  (loop npos (cons (substring/shared str pos npos) lst)))))))
    (if (zero? rem) sublst
      (append sublst (list (substring/shared str len end))) ) ) )

(define-syntax ensure-positive-fixnum
  (syntax-rules ()
    ((_ ?loc ?n) (set! ?n (inexact->exact (check-positive-fixnum ?loc ?n))) ) ) )

(define-inline (make-message-digest-primitive-lengths-list mdps)
	(map
	  message-digest-primitive-digest-length
    mdps) )

(define-inline (message-digest->integer-list str restriction)
	(map!
	  (lambda (str) (restriction (unsigned-integer32-ref str)))
    (string->substring-list/shared str unsigned-integer32-size)) )

(define-inline (bloom-filter-indicies-list bf obj)
	(map
	  (lambda (hasher) (hasher obj))
	  (%bloom-filter-hashes bf)) )

(define-inline (bloom-filter-fold bf proc init obj)
	(fold
	  proc
	  init
	  (take! (apply append! (bloom-filter-indicies-list bf obj)) (%bloom-filter-k bf))) )

;;; Calculators

;; Actual optimal: (expt (* n (log2 (/ m (- m 1)))) -1)
;; Returns the upper-bound, but w/ rounding

(define (optimum-k n m)
	(inexact->exact (round (* LN2 (/ m n)))) )

(define (optimum-m k n)
	(inexact->exact (round (/ (* n k) LN2))) )		; Similar to above

(define (p-random-one-bit k n m)
	(- 1 (expt (- 1 (/ 1 m)) (* k n))) )

(define (p-false-positive k n m)
	(expt (p-random-one-bit k n m) k) )

(define (desired-m p n . opt-k)
	(ensure-positive-fixnum 'desired-m n)
	(let loop ((m n))
		(let* ((k (optional opt-k (optimum-k n m)))
					 (calc-p (p-false-positive k n m)))
			(if (<= calc-p p) (values m k calc-p)
        (loop (fx+ m n))))) )

(define (actual-k mdps)
	(fold
		(lambda (len accum) (fx+ accum (fx/ len unsigned-integer32-size)))
		0
		(make-message-digest-primitive-lengths-list mdps)) )

;;; Bloom Filter

(define (bloom-filter? obj)
	(%bloom-filter? obj) )

(define-check+error-type bloom-filter)

(define (bloom-filter-n bf)
	(check-bloom-filter 'bloom-filter-n bf)
	(%bloom-filter-n bf) )

(define (bloom-filter-m bf)
	(check-bloom-filter 'bloom-filter-m bf)
	(%bloom-filter-m bf) )

(define (bloom-filter-k bf)
	(check-bloom-filter 'bloom-filter-k bf)
	(%bloom-filter-k bf) )

(define (make-bloom-filter m mdps #!optional des-k)
	(ensure-positive-fixnum 'make-bloom-filter m)
  (for-each (cut check-message-digest-primitive 'make-bloom-filter <>) mdps)
	(let ((make-mdp
          (lambda (mdp ranger)
            (lambda (obj)
              (message-digest->integer-list
                (message-digest-object mdp obj 'string)
                ranger) ) )))
		(let ((act-k (actual-k mdps)))
			(if (not des-k) (set! des-k act-k)
          (begin
            (ensure-positive-fixnum 'make-bloom-filter des-k)
            (when (fx< act-k des-k)
              (error 'make-bloom-filter
                "insufficient hash functions" act-k des-k) ) ) )
        ; Hash functions return integers,
        ; and m is a fixnum, so 'inexact->exact'
        ; will produce a fixnum.
			(%make-bloom-filter
			  0 m des-k
        (make-bit-vector m)
        (map
          (cut make-mdp <> (lambda (n) (inexact->exact (modulo n m))))
          mdps)))) )

(define (bloom-filter-p-false-positive bf . n)
	(check-bloom-filter 'bloom-filter-p-false-positive bf)
	(bloom-filter-p-false-positive
		(%bloom-filter-k bf)
		(optional n (%bloom-filter-n bf))
		(%bloom-filter-m bf)) )

(define (bloom-filter-set! bf obj)
	(check-bloom-filter 'bloom-filter-set! bf)
  (%bloom-filter-bits-set! bf
    (bloom-filter-fold bf
      (lambda (idx bits) (bit-vector-set! bits idx #t))
      (%bloom-filter-bits bf)
      obj))
	(%bloom-filter-n-set! bf (fx++ (%bloom-filter-n bf)))
	(void) )

(define (bloom-filter-exists? bf obj)
	(check-bloom-filter 'bloom-filter-exists? bf)
	(let ((bits (%bloom-filter-bits bf)))
	  (fx=
      (bloom-filter-fold bf
        (lambda (idx cnt) (if (bit-vector-ref bits idx) (fx++ cnt) cnt))
        0
        obj)
      (%bloom-filter-k bf)) ) )

) ;module bloom-filter
